// ImageMenu.cpp
//
// Copyright (c) 1997, Jim Barry

#include "stdafx.h"
#include "ImageMenu.h"

#include <map>
#include <set>
#include <algorithm>

// Local stuff
namespace
{
	OSVERSIONINFO const& GetVersion()
	{
		static OSVERSIONINFO info = { 0 };
		if (0 == info.dwOSVersionInfoSize)
		{
			info.dwOSVersionInfoSize = sizeof(info);
			GetVersionEx(&info);
		}
		return info;
	}

	bool MeasureAccelerators()
	{
		// WINBUG: Before we convert menu items to owner-draw, Win98 
		// calculates the largest accelerator width for the popup menu. 
		// It then adds this to the popup width after measuring the converted
		// owner-draw items!
		int nVer = 100 * GetVersion().dwMajorVersion + GetVersion().dwMinorVersion;
		return nVer != 410;
	}

	// Unfortunately, MaskBlt is not implemented on Win95, so we'll have to
	// do it our own way. See Ron Gery's "Bitmaps with Transparency" for 
	// more information.
	BOOL TransBlt(HDC hdcDest, int nXDest, int nYDest, int nWidth, int nHeight,
		HDC hdcSrc, int nXSrc, int nYSrc, HBITMAP hbmMask, int xMask, int yMask)
	{
		// Try it with MaskBlt first...
		const DWORD dwMagicRop = MAKEROP4(0x00AA0000, SRCCOPY);
		if (0 != MaskBlt(hdcDest, nXDest, nYDest, nWidth, nHeight,
		  hdcSrc, nXSrc, nYSrc, hbmMask, xMask, yMask, dwMagicRop))
			return TRUE;

		// Set up DCs and intermediate bitmap
		CDC& dcDest = *CDC::FromHandle(hdcDest);
		CDC& dcSrc = *CDC::FromHandle(hdcSrc);
		CDC dcImage; dcImage.CreateCompatibleDC(&dcDest);
		CDC dcMask; dcMask.CreateCompatibleDC(&dcDest);
		CBitmap& bmMask = *CBitmap::FromHandle(hbmMask);
		CBitmap bmImage; bmImage.CreateCompatibleBitmap(&dcDest, nWidth, nHeight);
		CBitmap* pBmImageOld = dcImage.SelectObject(&bmImage);
		
		// Copy destination to image
		dcImage.BitBlt(0, 0, nWidth, nHeight, &dcDest, nXDest, nYDest, SRCCOPY);

		// Invert source onto image
		dcImage.SetBkColor(dcDest.GetBkColor());
		dcImage.SetTextColor(dcDest.GetTextColor());
		dcImage.BitBlt(0, 0, nWidth, nHeight, &dcSrc, nXSrc, nYSrc, SRCINVERT);

		// Apply mask to image. We need to cater for the case where the 
		// mask is the same bitmap that is selected into the source DC
		// because a bitmap cannot be selected into two DCs at the same time.
		dcImage.SetBkColor(RGB(255, 255, 255));
		dcImage.SetTextColor(RGB(0, 0, 0));
		CBitmap* pBmSrcOld = dcSrc.SelectObject(pBmImageOld);
		CBitmap* pBmMaskOld = dcMask.SelectObject(&bmMask);
		dcImage.BitBlt(0, 0, nWidth, nHeight, &dcMask, xMask, yMask, SRCAND);
		dcMask.SelectObject(pBmMaskOld);
		dcSrc.SelectObject(pBmSrcOld);

		// Invert source onto image again
		dcImage.SetBkColor(dcDest.GetBkColor());
		dcImage.SetTextColor(dcDest.GetTextColor());
		dcImage.BitBlt(0, 0, nWidth, nHeight, &dcSrc, nXSrc, nYSrc, SRCINVERT);
		
		// Copy image to destination
		dcDest.BitBlt(nXDest, nYDest, nWidth, nHeight, &dcImage, 0, 0, SRCCOPY);

		// Clean up
		dcImage.SelectObject(pBmImageOld);
		return TRUE;
	}

	// Create a monochrome mask from a colour bitmap.
	HBITMAP CreateMask(CBitmap& bmSource, COLORREF crTrans)
	{
		// Create mono bitmap same size as source.
		BITMAP bm; 
		bmSource.GetBitmap(&bm);
		CBitmap bmMask;
		bmMask.CreateBitmap(bm.bmWidth, bm.bmHeight, 1, 1, 0);

		// Set up source and mask in memory DCs
		CDC dcMem1; dcMem1.CreateCompatibleDC(NULL);
		CDC dcMem2; dcMem2.CreateCompatibleDC(NULL);
		CBitmap* pbmOld1 = dcMem1.SelectObject(&bmSource);
		CBitmap* pbmOld2 = dcMem2.SelectObject(&bmMask);

		// Create mask
		dcMem1.SetBkColor(crTrans);
		dcMem2.BitBlt(0, 0, bm.bmWidth, bm.bmHeight, &dcMem1, 0, 0, SRCCOPY);

		// Clean up
		dcMem1.SelectObject(pbmOld1);
		dcMem2.SelectObject(pbmOld2);

		HBITMAP result = bmMask;
		bmMask.Detach();
		return result;
	}

	// Create a mono bitmap from a colour one, suitable for graying. 
	// The trick seems to be to set the 3D face and highlight colours
	// to 0 (background) and everything else to 1 (foreground).
	HBITMAP CreateMonoBitmap(CBitmap& bmSource)
	{
		// Create mono bitmap same size as source.
		BITMAP bm; 
		bmSource.GetBitmap(&bm);
		CBitmap bmMask;
		bmMask.CreateBitmap(bm.bmWidth, bm.bmHeight, 1, 1, 0);

		// Set up source and mask in memory DCs
		CDC dcMem1; dcMem1.CreateCompatibleDC(NULL);
		CDC dcMem2; dcMem2.CreateCompatibleDC(NULL);
		CBitmap* pbmOld1 = dcMem1.SelectObject(&bmSource);
		CBitmap* pbmOld2 = dcMem2.SelectObject(&bmMask);

		// Copy image, leaving COLOR_3DFACE behind.
		dcMem1.SetBkColor(GetSysColor(COLOR_3DFACE));
		dcMem2.BitBlt(0, 0, bm.bmWidth, bm.bmHeight, &dcMem1, 0, 0, SRCCOPY);

		// Erase COLOR_3DHIGHLIGHT from the mono bitmap.
		dcMem1.SetBkColor(GetSysColor(COLOR_3DHIGHLIGHT));
		dcMem2.BitBlt(0, 0, bm.bmWidth, bm.bmHeight, &dcMem1, 0, 0, SRCINVERT);

		// Clean up
		dcMem1.SelectObject(pbmOld1);
		dcMem2.SelectObject(pbmOld2);

		HBITMAP result = bmMask;
		bmMask.Detach();
		return result;
	}

	// Draw a "grayed" bitmap, using secret ROPs and black magic.
	// The source bitmap must be monochrome.
	BOOL DimBlt(HDC hdcDest, int nXDest, int nYDest, int nWidth, int nHeight,
		HBITMAP hbmMono, int xSrc, int ySrc)
	{
		// Select mono bitmap into off-screen dc.
		CDC dcMem; dcMem.Attach(CreateCompatibleDC(hdcDest));
		HGDIOBJ hbmOld = SelectObject(dcMem, hbmMono);

		// Draw highlight at (1,1) then shadow at (0,0).
		DWORD const rop = 0xb8074a; // "PSDPxax"
		COLORREF crOld = SetBkColor(hdcDest, RGB(255, 255, 255));
		HGDIOBJ hbrOld = SelectObject(hdcDest, GetSysColorBrush(COLOR_3DHIGHLIGHT));
		BitBlt(hdcDest, nXDest + 1, nYDest + 1, nWidth, nHeight, dcMem, xSrc, ySrc, rop);
		SelectObject(hdcDest, GetSysColorBrush(COLOR_3DSHADOW));
		BitBlt(hdcDest, nXDest, nYDest, nWidth, nHeight, dcMem, xSrc, ySrc, rop);

		// Clean up.
		SelectObject(hdcDest, hbrOld);
		SetBkColor(hdcDest, crOld);
		SelectObject(dcMem, hbmOld);
		return TRUE;
	}

	class Hooker
	{
	public:
		Hooker()
		{
		}
		
		~Hooker()
		{
			ASSERT(window_map.empty());
		}
		
		static void Register(CWnd* pWnd)
		{
			if (HWND hwnd = pWnd->GetSafeHwnd())
			{
				// Already registered?
				if (window_map.find(hwnd) != window_map.end())
					return;

				window_map[hwnd] = new Data(pWnd, 
					(WNDPROC)SetWindowLong(hwnd, 
						GWL_WNDPROC, (LONG)HookWndProc));
			}
		}

		static void Unregister(CWnd* pWnd)
		{
			if (HWND hwnd = pWnd->GetSafeHwnd())
			{
				// Is this window registered?
				if (window_map.find(hwnd) == window_map.end())
				{
					ASSERT(FALSE);
					return;
				}

				// Unsubclass window and remove map entry
				Data* pData = window_map[hwnd];
				if (HookWndProc != (WNDPROC)GetWindowLong(hwnd, GWL_WNDPROC))
					TRACE("Hooker::Unregister: Warning: window has been further subclassed\n");
				
				SetWindowLong(hwnd, 
					GWL_WNDPROC, (LONG)pData->pOldWndProc);
				window_map.erase(hwnd);
				delete pData;
			}
		}

		static LRESULT Default()
		{
			// MFC stores current MSG in thread state
			MSG& curMsg = AfxGetThreadState()->m_lastSentMsg;
			if (window_map.find(curMsg.hwnd) != window_map.end())
			{
				Data* pData = window_map[curMsg.hwnd];
				return ::CallWindowProc(pData->pOldWndProc, 
					curMsg.hwnd ,curMsg.message, curMsg.wParam, curMsg.lParam);
			}
			else
			{
				ASSERT(FALSE);
				return 0;
			}
		}

	private:
		static LRESULT CALLBACK
		HookWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
		{
#ifdef _USRDLL
			// If this is a DLL, need to set up MFC state
			AFX_MANAGE_STATE(AfxGetStaticModuleState());
#endif
			
			ASSERT(window_map.find(hwnd) != window_map.end());
			Data* pData = window_map[hwnd];

			static CImageMenu* pActiveMenu = NULL;
			MessageState state(hwnd, msg, wParam, lParam);

			LRESULT lr;
			switch (msg)
			{
			case WM_NCDESTROY:
				{
					// Window is being destroyed: unhook all hooks (for this window)
					// and pass msg to orginal window proc
					//
					WNDPROC wndproc = pData->pOldWndProc;
					Unregister(pData->pWndHooked);
					lr = ::CallWindowProc(wndproc, hwnd, msg, wParam, lParam);
					return lr;
				}
			case WM_SYSCOLORCHANGE:
			case WM_SETTINGCHANGE:
				{
					CImageMenu::Refresh();
					break;
				}
			case WM_INITMENU:
				{
					ASSERT(NULL == pActiveMenu);
					//TRACE("WM_INITMENU: %08x\n", pActiveMenu);
					if (HMENU hmenu = (HMENU)wParam)
					{
						if (CImageMenu* pMenu = CImageMenu::FromHandlePermanent(hmenu))
							pActiveMenu = pMenu->GetRoot();
					}
					break;
				}
			case WM_INITMENUPOPUP:
				{
					Default();
					HMENU hmenu = (HMENU)wParam;
					UINT uPos = (UINT) LOWORD(lParam);
					BOOL fSystemMenu = (BOOL)HIWORD(lParam);
					if (CImageMenu* pMenu = CImageMenu::FromHandlePermanent(hmenu))
					{
						CImageMenu* pRoot = pMenu->GetRoot();
						if (pActiveMenu != pRoot)
						{
							if (pActiveMenu)
								pActiveMenu->OnMenuClose();
							pActiveMenu = pRoot;
						}
						pMenu->OnInitMenuPopup(uPos, fSystemMenu);
					}
					return 0; // already called Default()
				}
			case WM_MEASUREITEM:
				{
					UINT idCtl = (UINT)wParam;// control identifier
					LPMEASUREITEMSTRUCT lpmis = (LPMEASUREITEMSTRUCT)lParam; // item-size information
					if (CImageMenu::OnMeasureItem(idCtl, lpmis))
						return TRUE;
					break;
				}
			case WM_DRAWITEM:
				{
					UINT idCtl = (UINT)wParam; // control identifier
					LPDRAWITEMSTRUCT lpdis = (LPDRAWITEMSTRUCT)lParam; // item-drawing information
					if (CImageMenu::OnDrawItem(idCtl, lpdis))
						return TRUE;
					break;
				}
			case WM_MENUCHAR:
				{
					TCHAR chUser = (TCHAR)LOWORD(wParam); // character code
					UINT fuFlag = (UINT)HIWORD(wParam); // menu flag 
					HMENU hmenu = (HMENU)lParam; // handle of menu 
					if (CImageMenu* pMenu = CImageMenu::FromHandlePermanent(hmenu))
						return CImageMenu::OnMenuChar(chUser, fuFlag, pMenu);
					break;
				}
			case WM_MENUSELECT:
				{
					UINT fuFlag = (UINT)HIWORD(wParam); // menu flag 
					HMENU hmenu = (HMENU)lParam; // handle of menu 
					if (0xffff == fuFlag && NULL == hmenu) // menu closed
					{
						if (pActiveMenu)
						{
							pActiveMenu->OnMenuClose();
							pActiveMenu = NULL;
						}
					}
					break;
				}
			case WM_COMMAND:
				{
					// Sadly, OLE containers do not foward WM_ENTERMENULOOP or 
					// WM_EXITMENULOOP, but we can catch accelerators by looking
					// for WM_COMMANDs with a notify code of 1.
					WORD wNotifyCode = HIWORD(wParam);
					if (1 == wNotifyCode)
					{
						if (pActiveMenu)
						{
							pActiveMenu->OnMenuClose();
							pActiveMenu = NULL;
						}
					}
					break;
				}
			}

			return Default();
		}

		struct MessageState
		{
			MSG msgSave;
			MessageState(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
			{
				MSG& msgCur = AfxGetThreadState()->m_lastSentMsg;
				msgSave = msgCur;
				msgCur.hwnd = hwnd;
				msgCur.message = msg;
				msgCur.wParam = wp;
				msgCur.lParam = lp;
			}
			~MessageState()
			{
				AfxGetThreadState()->m_lastSentMsg = msgSave;
			}
		};

		struct Data
		{
			Data(CWnd* pWnd, WNDPROC pWndProc) {
				pWndHooked = pWnd;
				pOldWndProc = pWndProc;
				pActiveMenu = NULL;
			}
			CWnd* pWndHooked;
			WNDPROC pOldWndProc; // ..and original window proc
			CImageMenu* pActiveMenu; // Current menu
		};
		typedef std::map<HWND, Data*> WindowMap;
		static WindowMap window_map;
	};
	// Definition of static members
	Hooker::WindowMap Hooker::window_map;

	// The window hooker
	Hooker hooker;

	// All CImageMenu objects enter themselves in here
	typedef std::set<CImageMenu*> MenuMap;
	MenuMap menu_map;

	// Metrics
	const int g_cxTextIndent = 2;
	const int g_cxIconSpace = 2;
	const int g_cyIconSpace = 2;
}

// Static members
CFont CImageMenu::s_fontMenu;
CFont CImageMenu::s_fontMenuBold;

CImageMenu* CImageMenu::FromHandlePermanent(HMENU hMenu)
{
	CMenu* pMenu = CMenu::FromHandlePermanent(hMenu);
	return DYNAMIC_DOWNCAST(CImageMenu, pMenu);
}

IMPLEMENT_DYNCREATE(CImageMenu, CMenu)

CImageMenu::CImageMenu()
{
	m_pOwner = NULL;
	m_nImageWidth = 0;
	m_nArrowWidth = 0;
	m_flags = show_images | destroy_hmenu;
	menu_map.insert(this);
}

CImageMenu::CImageMenu(CImageMenu* pOwner)
{
	m_pOwner = pOwner;
	m_nImageWidth = 0;
	m_nArrowWidth = 0;
	m_flags = show_images | popup_menu | destroy_hmenu;
	menu_map.insert(this);
}

CImageMenu* CImageMenu::GetRoot()
{
	CImageMenu* pMenu = this;
	while (pMenu->m_pOwner)
		pMenu = pMenu->m_pOwner;
	return pMenu;
}

void CImageMenu::Register(CWnd* pWnd)
{
	hooker.Register(pWnd);
}

void CImageMenu::Refresh()
{
	// Reload the menu fonts
	s_fontMenu.DeleteObject();
	GetFont();
	s_fontMenuBold.DeleteObject();
	GetBoldFont();

	// Call OnRefresh for all the CImageMenus
	std::for_each(menu_map.begin(), menu_map.end(), 
		std::mem_fun(OnRefresh));
}

BOOL CImageMenu::OnRefresh()
{
	return TRUE;
}

CFont* CImageMenu::GetFont()
{
	// Lazy creation
	if (NULL == s_fontMenu.GetSafeHandle())
	{
		NONCLIENTMETRICS ncm;
		ncm.cbSize = sizeof(ncm);
		SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, &ncm, 0);
		s_fontMenu.CreateFontIndirect(&ncm.lfMenuFont);
	}
	return &s_fontMenu;
}

CFont* CImageMenu::GetBoldFont()
{
	// Lazy creation
	if (NULL == s_fontMenuBold.GetSafeHandle())
	{
		NONCLIENTMETRICS ncm;
		ncm.cbSize = sizeof(ncm);
		SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, &ncm, 0);
		ncm.lfMenuFont.lfWeight = 700;
		s_fontMenuBold.CreateFontIndirect(&ncm.lfMenuFont);
	}
	return &s_fontMenuBold;
}

BOOL CImageMenu::GetFlag(Flag flag)
{
	return m_flags & flag;
}

void CImageMenu::SetFlag(Flag flag, BOOL fSet)
{
	if (fSet)
		m_flags |= flag;
	else
		m_flags &= ~flag;
}

BOOL CImageMenu::IsPopupMenu()
{
	return m_flags & popup_menu;
}

BOOL CImageMenu::CreatePopupMenu(CImageMenu* pOwner, BOOL fAutoDelete)
{
	if (!CMenu::CreatePopupMenu())
		return FALSE;

	m_pOwner = pOwner;
	m_flags |= popup_menu;

	// Add to submenu list
	if (m_pOwner)
	{
		if (fAutoDelete)
			m_pOwner->m_listSubMenus.insert(this);
	}
	else
		// Can't autodelete if not owned
		ASSERT(!fAutoDelete);

	return TRUE;
}

CImageMenu* CImageMenu::GetSubMenu(int nPos)
{
	if (HMENU hmenu = ::GetSubMenu(GetSafeHmenu(), nPos))
	{
		if (CImageMenu* pMenu = FromHandlePermanent(hmenu))
			return pMenu;
	}
	return NULL;
}

BOOL CImageMenu::Attach(HMENU hMenu, BOOL fIsPopupMenu)
{
	if (!CMenu::Attach(hMenu))
		return FALSE;

	SetFlag(popup_menu, fIsPopupMenu);
	if (!fIsPopupMenu)
		ProcessMenuItems(FALSE);

	return TRUE;
}

void CImageMenu::ShowImages(BOOL fShow)
{
	SetFlag(show_images, fShow);
}

BOOL CImageMenu::ProcessMenuItem(UINT nItem, BOOL fShowImages)
{
	MENUITEMINFO mii;
	ZeroMemory(&mii, sizeof(mii));
	mii.cbSize = sizeof(mii);
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID | MIIM_SUBMENU | MIIM_DATA;
	if (!GetMenuItemInfo(nItem, TRUE, &mii))
		return FALSE;

	// States to ignore
	const UINT fuIgnore = MFT_BITMAP;

	// Is it a popup?
	if (mii.hSubMenu)
	{
		if (NULL == FromHandlePermanent(mii.hSubMenu))
		{
			// Leave other people's owner-draw submenu items alone.
			if (mii.fType & MFT_OWNERDRAW)
				return TRUE;

			// Create a new CImageMenu object for the submenu
			CImageMenu* pNewMenu = (CImageMenu*)GetRuntimeClass()->CreateObject();
			pNewMenu->m_pOwner = this;
			pNewMenu->SetFlag(destroy_hmenu, FALSE);
			pNewMenu->SetFlag(show_images, GetFlag(show_images));
			pNewMenu->Attach(mii.hSubMenu, TRUE);
			m_listSubMenus.insert(pNewMenu);
		}
	}

	if (0 == (mii.fType & fuIgnore))
	{
		// Is it one of ours?
		if (CItemData* pData = GetItemData(mii.dwItemData))
		{
			if (fShowImages)
			{
				// If it is still owner-draw then leave it alone.
				if (mii.fType & MFT_OWNERDRAW)
					return TRUE;

				// Somebody modified the item, so delete the data and fall through.
				mii.dwItemData = 0;
				DestroyItemData(pData);
			}
			else
			{
				// Return menu to original state
				mii.dwTypeData = pData->m_strText.GetBuffer(0);
				mii.fType &= ~MFT_OWNERDRAW;
				mii.dwItemData = 0; // probably unnecessary
				RemoveMenu(nItem, MF_BYPOSITION);
				VERIFY(InsertMenuItem(nItem, TRUE, &mii));
				DestroyItemData(pData);
				TRACE("Unconverted menu item ID %d\n", mii.wID);
				return TRUE; // Done
			}
		}
		else
		{
			// Leave other people's owner-draw items alone.
			if (mii.fType & MFT_OWNERDRAW)
				return TRUE;
		}

		if (fShowImages && GetFlag(popup_menu))
		{
			CString strLabel;
			GetMenuString(nItem, strLabel, MF_BYPOSITION);
			CItemData* pData = CreateItemData(strLabel);
			mii.fType |= MFT_OWNERDRAW;
			mii.dwItemData = (DWORD)pData;
			mii.dwTypeData = 0;
			mii.fMask = MIIM_TYPE | MIIM_DATA;
			VERIFY(SetMenuItemInfo(nItem, TRUE, &mii));
			TRACE("Converted menu item ID %d\n", mii.wID);
		}
	}
	else
	{
		TRACE("CImageMenu::ProcessMenuItem: Item #%u ignored\n", nItem);
	}

	return TRUE;
}

BOOL CImageMenu::ProcessMenuItems(BOOL fShowImages)
{
	// Add menu bar breaks if we exceed the screen height
	const int nMaxHeight = 
		GetSystemMetrics(SM_CYSCREEN) - 2 * GetSystemMetrics(SM_CYEDGE);
	int nHeight = 0;
	MEASUREITEMSTRUCT mis;
	ZeroMemory(&mis, sizeof(mis));
	mis.CtlType = ODT_MENU;
	MENUITEMINFO mii;
	
	UINT nCount = GetMenuItemCount();
	for (UINT nItem = 0; nItem < nCount; nItem++)
	{
		ProcessMenuItem(nItem, fShowImages);
		if (fShowImages)
		{
			// Measure the item and see if the menu exceeds the screen height
			ZeroMemory(&mii, sizeof(mii));
			mii.cbSize = sizeof(mii);
			mii.fMask = MIIM_TYPE|MIIM_STATE|MIIM_ID|MIIM_SUBMENU|MIIM_DATA;
			VERIFY(GetMenuItemInfo(nItem, TRUE, &mii));
			mis.itemID = mii.wID;
			mis.itemData = mii.dwItemData;
			MeasureItem(&mis);
			nHeight += mis.itemHeight;
			if (nHeight > nMaxHeight)
			{
				// Too big! Add a column bar break
				mii.fType |= MFT_MENUBARBREAK;
				SetMenuItemInfo(nItem, TRUE, &mii);
				nHeight = mis.itemHeight;
			}
		}
	}

	// These will be recalculated when items are measured
	if (!fShowImages)
	{
		m_nImageWidth = 0;
		m_nArrowWidth = 0;
	}

	return TRUE;
}

BOOL CImageMenu::LoadMenu(LPCTSTR lpszResourceName)
{
	if (!CMenu::LoadMenu(lpszResourceName))
		return FALSE;

	return ProcessMenuItems(FALSE);
}

BOOL CImageMenu::LoadMenu(UINT nIDResource)
{
	if (!CMenu::LoadMenu(nIDResource))
		return FALSE;

	return ProcessMenuItems(FALSE);
}


void CImageMenu::AdjustABCWidths(CDC& dc, const CString& str, CRect& rect, BOOL fShrink)
{
	// DrawText with DT_CALCRECT does not allow for overhang on italic fonts.
	// Correct this by adding in the 'C' of the ABC width of the last character.
	// This only seems to matter for TrueType fonts so we don't bother with
	// GetCharABCWidthsFloat.
	if (str.GetLength() > 0)
	{
		// Adjust trailing edge
		ABC abc;
		TCHAR ch = str[str.GetLength() - 1];
		if (dc.GetCharABCWidths(ch, ch, &abc))
			rect.right -= fShrink ? abs(abc.abcC) : -abs(abc.abcC);
	}
}

void CImageMenu::DrawMenuText(CDC& dc, const CString& strItem, const CRect& rect)
{
	const UINT nFormat = DT_SINGLELINE | DT_VCENTER | DT_NOCLIP;
	CRect r(rect);

	// Compensate for italic overhang
	AdjustABCWidths(dc, strItem, r, TRUE);

	// Does menu item contain a tab char?
	int nTabIndex = strItem.Find(_T('\t'));
	if (nTabIndex >= 0)
	{
		CString str1(strItem.Left(nTabIndex)); // Command text
		CString str2(strItem.Mid(nTabIndex+1)); // Shortcut text
		dc.DrawText(str1, r, nFormat);
		dc.DrawText(str2, r, nFormat|DT_RIGHT);
	}
	else
		dc.DrawText(strItem, r, nFormat);
}

void CImageMenu::DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	// Copy the DRAWITEMSTRUCT so we can mess around with it
	DRAWITEMSTRUCT dis = *lpDrawItemStruct;
	ASSERT(ODT_MENU == dis.CtlType);
	CDC& dc = *CDC::FromHandle(dis.hDC);
	UINT& itemID = dis.itemID;
	CRect& rcItem = (CRect&)dis.rcItem;

	// Is it a separator?
	MENUITEMINFO mii;
	ZeroMemory(&mii, sizeof(mii));
	mii.cbSize = sizeof(mii);
	mii.fMask = MIIM_TYPE;
	VERIFY(GetMenuItemInfo(itemID, FALSE, &mii));
	if (MFT_SEPARATOR & mii.fType)
	{
		DrawSeparator(&dis);
	}
	else
	{
		// Use double buffering
		CDC dcMem; 
		CBitmap bmBack; 
		CBitmap* pOldBitmap = 0;
		CRect rcItemInit = rcItem;
		if (dcMem.CreateCompatibleDC(&dc) &&
			bmBack.CreateCompatibleBitmap(&dc, rcItem.Width(), rcItem.Height()) &&
			0 != (pOldBitmap = dcMem.SelectObject(&bmBack)))
		{
			dis.hDC = dcMem;
			rcItemInit = CRect(CPoint(0, 0), rcItem.Size());
		}

		// Draw image
		rcItem = rcItemInit;
		rcItem.right = rcItem.left + m_nImageWidth;
		DrawImage(&dis);

		// Draw text
		rcItem = rcItemInit;
		rcItem.left = rcItem.left + m_nImageWidth;
		rcItem.right = rcItem.right - m_nArrowWidth;
		DrawText(&dis);

		// Draw submenu arrow
		rcItem = rcItemInit;
		rcItem.left = rcItem.right - m_nArrowWidth;
		DrawArrow(&dis);

		// Blit the back buffer to the screen
		if (pOldBitmap)
		{
			CRect r = lpDrawItemStruct->rcItem;
			dc.BitBlt(r.left, r.top, r.Width(), r.Height(), &dcMem, 0, 0, SRCCOPY);
			dcMem.SelectObject(pOldBitmap);
		}
	}

	// WINBUG: Stop Windows from drawing submenu arrows (or anything else)
	// over the top of this item.
	dc.ExcludeClipRect(&lpDrawItemStruct->rcItem);
}

void CImageMenu::MeasureItem(LPMEASUREITEMSTRUCT lpMeasureItemStruct)
{
	const UINT nFormat = DT_SINGLELINE|DT_CALCRECT;
	UINT& itemWidth = lpMeasureItemStruct->itemWidth;
	UINT& itemHeight = lpMeasureItemStruct->itemHeight;
	UINT& itemID = lpMeasureItemStruct->itemID;
	DWORD& itemData = lpMeasureItemStruct->itemData;
	CSize size;

	TRACE("Measuring item ID %d\n", itemID);

	// Get the menu item type.
	MENUITEMINFO mii;
	ZeroMemory(&mii, sizeof(mii));
	mii.cbSize = sizeof(mii);
	mii.fMask = MIIM_TYPE;
	VERIFY(GetMenuItemInfo(itemID, FALSE, &mii));

	// Is it a separator?
	if (MFT_SEPARATOR & mii.fType)
	{
		// Set both width and height - the struct seems sometimes 
		// to arrive containing garbage in the width.
		size = MeasureSeparator(lpMeasureItemStruct);
		itemWidth = size.cx;
		itemHeight = size.cy;
	}
	else
	{
		// Measure image
		size = MeasureImage(lpMeasureItemStruct);
		m_nImageWidth = max(m_nImageWidth, size.cx);
		itemWidth = m_nImageWidth;
		itemHeight = max(itemHeight, (UINT)size.cy);

		// Measure text
		size = MeasureText(lpMeasureItemStruct);
		itemWidth += size.cx;
		itemHeight = max(itemHeight, (UINT)size.cy);

		// Measure submenu arrow
		size = MeasureArrow(lpMeasureItemStruct);
		m_nArrowWidth = max(m_nArrowWidth, size.cx);
		itemWidth += m_nArrowWidth;
		itemHeight = max(itemHeight, (UINT)size.cy);

		// Windows adds the checkmark to the width, so compensate.
		itemWidth -= GetSystemMetrics(SM_CXMENUCHECK);
	}
}

CSize CImageMenu::MeasureSeparator(LPMEASUREITEMSTRUCT lpMeasureItemStruct)
{
	// Only the height is relevant
	return CSize(0, 10); // looks OK
}

CSize CImageMenu::MeasureImage(LPMEASUREITEMSTRUCT lpMeasureItemStruct)
{
	TRACE("CImageMenu::MeasureImage: itemID = %x\n", lpMeasureItemStruct->itemID);
	// Implement in derived class
	return CSize(0, 0);
}

CSize CImageMenu::MeasureText(LPMEASUREITEMSTRUCT lpMeasureItemStruct)
{
	const UINT nFormat = DT_SINGLELINE|DT_CALCRECT;
	CString strItem = GetTextFromItemData(lpMeasureItemStruct->itemData);
	UINT itemID = lpMeasureItemStruct->itemID;
	BOOL fDefault = (GetMenuDefaultItem(FALSE) == itemID);

	// Set up DC
	CClientDC dc(NULL);
	CRect rect(0, 0, 0, 0);
	dc.SelectObject(fDefault ? GetBoldFont() : GetFont());

	// Does menu item contain a tab char?
	int nTabIndex = strItem.Find('\t');
	if (nTabIndex >= 0)
	{
		CString str1(strItem.Left(nTabIndex));
		CString str2(strItem.Mid(nTabIndex+1));
		CRect rect2(0, 0, 0, 0);
		dc.DrawText(str1, rect, nFormat);
		dc.DrawText(str2, rect2, nFormat);
		if (MeasureAccelerators())
			rect.right += 20 + rect2.Width();
	}
	else
	{
		dc.DrawText(strItem, rect, nFormat);
	}
	rect.right += 2 * g_cxTextIndent; // left and right indent
	rect.bottom++; // looks better

	// Allow for italic overhang
	AdjustABCWidths(dc, strItem, rect, FALSE);

	return rect.Size();
}

CSize CImageMenu::MeasureArrow(LPMEASUREITEMSTRUCT lpMeasureItemStruct)
{
	int nWidth = GetSystemMetrics(SM_CXMENUCHECK);
	int nHeight = GetSystemMetrics(SM_CYMENUCHECK);
	return CSize(nWidth, nHeight);
}

void CImageMenu::DrawSeparator(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	CDC& dc = *CDC::FromHandle(lpDrawItemStruct->hDC);
	CRect rcItem(lpDrawItemStruct->rcItem);

	CRect rcSep(rcItem);
	rcSep.top = rcSep.bottom = (rcItem.top + rcItem.bottom) / 2;
	rcSep.InflateRect(-14, 1);
	FillRect(dc, rcItem, GetSysColorBrush(COLOR_3DFACE));
	dc.DrawEdge(rcSep, BDR_SUNKENOUTER, BF_TOP|BF_BOTTOM);
}

void CImageMenu::DrawImage(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	// Implement in derived class
}

void CImageMenu::DrawText(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	CDC& dc = *CDC::FromHandle(lpDrawItemStruct->hDC);
	CRect rcText(lpDrawItemStruct->rcItem);
	UINT itemID = lpDrawItemStruct->itemID;
	CString strItem = GetTextFromItemData(lpDrawItemStruct->itemData);
	BOOL fSelected = lpDrawItemStruct->itemState & ODS_SELECTED;
	BOOL fChecked = lpDrawItemStruct->itemState & ODS_CHECKED;
	BOOL fGrayed = lpDrawItemStruct->itemState & ODS_GRAYED;
	BOOL fDefault = (GetMenuDefaultItem(FALSE) == itemID);

	int nBackground = fSelected ? COLOR_HIGHLIGHT : COLOR_3DFACE;
	int nText = fSelected ? COLOR_HIGHLIGHTTEXT : COLOR_MENUTEXT;

	// Draw the text
	FillRect(dc, rcText, GetSysColorBrush(nBackground));
	rcText.DeflateRect(g_cxTextIndent, 0);
	dc.SelectObject(fDefault ? GetBoldFont() : GetFont());
	dc.SelectStockObject(NULL_BRUSH);
	dc.SetBkMode(TRANSPARENT);
	if (fGrayed)
	{
		if (!fSelected)
		{
			CRect rcText2(rcText); rcText2.OffsetRect(1, 1);
			dc.SetTextColor(GetSysColor(COLOR_3DHIGHLIGHT));
			DrawMenuText(dc, strItem, rcText2);
			dc.SetTextColor(GetSysColor(COLOR_3DSHADOW));
		}
		else
		{
			// WINBUG: sometimes the highlight colour and graytext colour
			// are the same!? If so, just use the regular text colour.
			if (GetSysColor(COLOR_HIGHLIGHT) != GetSysColor(COLOR_GRAYTEXT))
				dc.SetTextColor(GetSysColor(COLOR_GRAYTEXT));
			else
				dc.SetTextColor(GetSysColor(COLOR_MENUTEXT));
		}
	}
	else
	{
		dc.SetTextColor(GetSysColor(nText));
	}
	DrawMenuText(dc, strItem, rcText);
}

void CImageMenu::DrawArrow(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	CDC& dc = *CDC::FromHandle(lpDrawItemStruct->hDC);
	CRect rcArrow(lpDrawItemStruct->rcItem);
	UINT itemID = lpDrawItemStruct->itemID;
	BOOL fSelected = lpDrawItemStruct->itemState & ODS_SELECTED;
	BOOL fGrayed = lpDrawItemStruct->itemState & ODS_GRAYED;

	int nBackground = fSelected ? COLOR_HIGHLIGHT : COLOR_3DFACE;
	int nText = fSelected ? COLOR_HIGHLIGHTTEXT : COLOR_MENUTEXT;
	// WINBUG: sometimes the highlight colour and graytext colour
	// are the same!? If so, just use the regular text colour.
	if (fGrayed && fSelected)
	{
		if (GetSysColor(COLOR_HIGHLIGHT) != GetSysColor(COLOR_GRAYTEXT))
			nText = COLOR_GRAYTEXT;
		else
			nText = COLOR_MENUTEXT;
	}
	dc.SetTextColor(GetSysColor(nText));
	dc.SetBkColor(GetSysColor(nBackground));

	// Fill in background
	FillRect(dc, rcArrow, GetSysColorBrush(nBackground));

	// Draw the submenu arrow
	if (MF_POPUP & GetMenuState(itemID, MF_BYCOMMAND))
	{
		CRect& r = rcArrow; // save some typing
		CDC dcMem; 
		dcMem.CreateCompatibleDC(NULL);
		CBitmap bm; 
		bm.CreateCompatibleBitmap(&dcMem, r.Width(), r.Height());
		CBitmap* pbmOld = dcMem.SelectObject(&bm);
		CRect rectDFC(0, 0, r.Width(), r.Height());
		FillRect(dcMem, rectDFC, (HBRUSH)GetStockObject(WHITE_BRUSH));
		dcMem.DrawFrameControl(rectDFC, DFC_MENU, DFCS_MENUARROW);
		if (fGrayed && !fSelected)
		{
			dc.SetTextColor(GetSysColor(COLOR_3DHIGHLIGHT));
			TransBlt(dc, r.left + 1, r.top + 1, r.Width(), r.Height(), 
				dcMem, 0, 0, bm, 0, 0);
			dc.SetTextColor(GetSysColor(COLOR_3DSHADOW));
		}
		TransBlt(dc, r.left, r.top, r.Width(), r.Height(), 
			dcMem, 0, 0, bm, 0, 0);
		dcMem.SelectObject(pbmOld);
	}
}

CImageMenu::CItemData* CImageMenu::CreateItemData(LPCTSTR lpItemText)
{
	CItemData* pData = new CItemData(this, lpItemText);
	m_listItemData.insert(pData);
	return pData;
}

void CImageMenu::DestroyItemData(CItemData* pData)
{
	m_listItemData.erase(pData);
	delete pData;
}

CImageMenu::CItemData* CImageMenu::GetItemData(DWORD dwItemData)
{
	if (-1 == dwItemData) // separator
		return 0;
	
	if (CItemData* p = reinterpret_cast<CItemData*>(dwItemData))
		if (AfxIsValidAddress(p, sizeof(*p)))
			if (CItemData::magic_number == p->m_id)
				return p;

	return 0; // Naaaah
}

CString CImageMenu::GetTextFromItemData(DWORD dwItemData)
{
	CItemData* pData = GetItemData(dwItemData);
	return pData ? pData->m_strText : "";
}

BOOL CImageMenu::DestroyMenu()
{
	// Delete submenus
	{
		int count = m_listSubMenus.size();
		if (count > 0)
			TRACE("CImageMenu::DestroyMenu: deleting %d submenu(s)\n", count);
		for (SubMenuPool::iterator it = m_listSubMenus.begin(); 
			it != m_listSubMenus.end(); ++it)
				delete *it;
		m_listSubMenus.clear();
	}

	// Delete labels
	{
		int count = m_listItemData.size();
		if (count > 0)
			TRACE("CImageMenu::DestroyMenu: deleting %d data item(s)\n", count);
		for (ItemDataPool::iterator it = m_listItemData.begin(); 
			it != m_listItemData.end(); ++it)
				delete *it;
		m_listItemData.clear();
	}

	if (!GetFlag(destroy_hmenu))
	{
		Detach();
		return TRUE;
	}
	return CMenu::DestroyMenu();
}

CImageMenu::~CImageMenu()
{
	menu_map.erase(this);
	DestroyMenu();
}

BOOL CImageMenu::FindMnemonic(CString str, UINT nChar)
{
	str.MakeUpper(); // str is by-value
	int nFind = str.Find(CString(_T('&')) + CString(toupper(nChar)));
	if ((nFind > 0) && (str[nFind-1] == _T('&'))) // check for double-ampersand
		return FALSE;
	return nFind >= 0;
}

LRESULT CImageMenu::OnMenuChar(UINT nChar, UINT nFlags, CMenu* pMenu)
{
	CImageMenu* pImageMenu = DYNAMIC_DOWNCAST(CImageMenu, pMenu);
	if (pImageMenu)
		return pImageMenu->OnMenuChar(nChar, nFlags);

	return 0;
}

LRESULT CImageMenu::OnMenuChar(UINT nChar, UINT nFlags)
{
	const UINT nInvalid = -1;
	MENUITEMINFO mii;

	// Check we have a menu with at least one item!
	UINT nMenuCount = GetMenuItemCount();
	if (nInvalid == nMenuCount || 0 == nMenuCount)
		return MAKELRESULT(0, MNC_IGNORE);

	// Some working variables
	UINT nItem = 0;
	UINT nItemsFound = 0;
	UINT nFirstFind = nInvalid;
	BOOL fFoundSelection = FALSE; 

	// Scan menu items
	for (nItem = 0; nItem < nMenuCount; nItem++)
	{
		ZeroMemory(&mii, sizeof(mii));
		mii.cbSize = sizeof(mii);
		mii.fMask = MIIM_TYPE|MIIM_DATA|MIIM_STATE;

		VERIFY(GetMenuItemInfo(nItem, TRUE, &mii));
		if (mii.fType & MFT_OWNERDRAW)
		{
			CString strItem = GetTextFromItemData(mii.dwItemData);
			if (FindMnemonic(strItem, nChar))
			{
				nItemsFound++;
				if ((nFirstFind == nInvalid) || fFoundSelection)
				{
					nFirstFind = nItem;
					fFoundSelection = FALSE;
				}
			}
		}

		if (mii.fState & MFS_HILITE)
			fFoundSelection = TRUE;
	}

	// Did we find the mnemonic?
	if (0 == nItemsFound)
		// no match
		return MAKELRESULT(0, MNC_IGNORE);

	if (1 == nItemsFound)
		// one match - excecute the command
		return MAKELRESULT(nFirstFind, MNC_EXECUTE);

	// more than one match - select the first match
	return MAKELRESULT(nFirstFind, MNC_SELECT);
}

void CImageMenu::OnInitMenuPopup(CMenu* pPopupMenu, UINT nIndex, BOOL bSysMenu)
{
	CImageMenu* pImageMenu = DYNAMIC_DOWNCAST(CImageMenu, pPopupMenu);
	if (pImageMenu)
		pImageMenu->OnInitMenuPopup(nIndex, bSysMenu);
}

void CImageMenu::OnInitMenuPopup(UINT nIndex, BOOL bSysMenu)
{
	ProcessMenuItems(GetFlag(show_images));
}

void CImageMenu::OnMenuSelect(UINT uItem, UINT fuFlags, HMENU hmenu)
{
	if (NULL == hmenu && 0xFFFF == fuFlags)
		OnMenuClose();
}

void CImageMenu::OnMenuClose()
{
	for (SubMenuPool::iterator it = m_listSubMenus.begin(); 
		it != m_listSubMenus.end(); ++it)
	{
		(*it)->OnMenuClose();
	}
	ProcessMenuItems(FALSE);
}

BOOL CImageMenu::OnMeasureItem(int nIDCtl, LPMEASUREITEMSTRUCT lpMeasureItemStruct)
{
	BOOL result = FALSE;
	UINT& CtlType = lpMeasureItemStruct->CtlType;
	DWORD& itemData = lpMeasureItemStruct->itemData;

	// Dispatch to correct menu object
	if (ODT_MENU == CtlType)
	{
		CItemData* pData = GetItemData(itemData);
		if (pData && pData->m_pMenu)
		{
			pData->m_pMenu->MeasureItem(lpMeasureItemStruct);
			result = TRUE;
		}
	}
	return result;
}

BOOL CImageMenu::OnDrawItem(int nIDCtl, LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	UINT& CtlType = lpDrawItemStruct->CtlType;
	DWORD& itemData = lpDrawItemStruct->itemData;

	// Dispatch to correct menu object
	if (ODT_MENU == CtlType)
	{
		CItemData* pData = GetItemData(itemData);
		if (pData && pData->m_pMenu)
		{
			pData->m_pMenu->DrawItem(lpDrawItemStruct);
			return TRUE;
		}
	}
	return FALSE;
}

//
// CCommandMenu
//

namespace
{
	// Ripped off from MFC...
	struct CToolBarData
	{
		WORD wVersion;
		WORD wWidth;
		WORD wHeight;
		WORD wItemCount;
		//WORD aItems[wItemCount]

		WORD* items()
			{ return (WORD*)(this+1); }
	};
}

void CCommandMenu::ToolBarInfo::Reload()
{
	// Load the bitmaps
	bmNormal.DeleteObject();
	bmNormal.Attach(
		AfxLoadSysColorBitmap(hInstImageWell, hRsrcImageWell));

	bmMask.DeleteObject();
	bmMask.Attach(CreateMask(bmNormal, GetSysColor(COLOR_3DFACE)));

	bmMono.DeleteObject();
	bmMono.Attach(CreateMonoBitmap(bmNormal));
}

// Static members
BOOL CCommandMenu::s_fShowButtons = TRUE;

IMPLEMENT_DYNCREATE(CCommandMenu, CImageMenu)

CCommandMenu::CCommandMenu()
{
	m_sizeImage = CSize(0, 0);
}

BOOL CCommandMenu::AddToolbar(UINT nIDResource)
{
	return AddToolbar(MAKEINTRESOURCE(nIDResource));
}

BOOL CCommandMenu::AddToolbars(const UINT* id_list, UINT count)
{
	for (UINT i = 0; i < count; ++i)
		if (!AddToolbar(id_list[i]))
			return FALSE;
	return TRUE;
}

BOOL CCommandMenu::AddToolbar(LPCTSTR lpszResourceName)
{
	// determine location of the bitmap in resource fork
	HINSTANCE hInst = AfxFindResourceHandle(lpszResourceName, RT_TOOLBAR);
	HRSRC hRsrc = ::FindResource(hInst, lpszResourceName, RT_TOOLBAR);
	if (hRsrc == NULL)
	{
		TRACE("CCommandMenu::AddToolBar: Cannot find toolbar resource\n");
		return FALSE;
	}

	HGLOBAL hGlobal = LoadResource(hInst, hRsrc);
	if (hGlobal == NULL)
		return FALSE;

	CToolBarData* pData = (CToolBarData*)LockResource(hGlobal);
	if (pData == NULL)
		return FALSE;
	ASSERT(pData->wVersion == 1);

	// Add a new ToolBarInfo to the list
	ToolBarInfo* pTBI = new ToolBarInfo;
	m_listToolBars.AddTail(pTBI);

	// Get the button IDs
	for (int i = 0; i < pData->wItemCount; i++)
	{
		// Filter out separators (zero IDs)
		UINT nID = pData->items()[i];
		if (nID > 0)
			pTBI->nItemIDs.Add(nID);
	}

	// Get the button size
	pTBI->sizeImage = CSize(pData->wWidth, pData->wHeight);
	m_sizeImage.cx = max(GetImageSize().cx, pTBI->sizeImage.cx);
	m_sizeImage.cy = max(GetImageSize().cy, pTBI->sizeImage.cy);

	// Find the corresponding bitmap resource
	pTBI->hInstImageWell = AfxFindResourceHandle(lpszResourceName, RT_BITMAP);
	pTBI->hRsrcImageWell = ::FindResource(pTBI->hInstImageWell, lpszResourceName, RT_BITMAP);

	// Load the bitmap
	pTBI->Reload();

	UnlockResource(hGlobal);
	FreeResource(hGlobal);

	return TRUE;
}

CSize CCommandMenu::GetImageSize()
{
	// Go to the top!
	CCommandMenu* pOwner = DYNAMIC_DOWNCAST(CCommandMenu, m_pOwner);
	if (pOwner)
		return pOwner->GetImageSize();

	return m_sizeImage;
}

CCommandMenu::ToolBarInfo* CCommandMenu::FindToolBar(UINT nID)
{
	// Go to the top!
	CCommandMenu* pOwner = DYNAMIC_DOWNCAST(CCommandMenu, m_pOwner);
	if (pOwner)
		return pOwner->FindToolBar(nID);

	POSITION pos = m_listToolBars.GetHeadPosition();
	while (pos)
	{
		ToolBarInfo* pTBI = m_listToolBars.GetNext(pos);
		const int nCount = pTBI->nItemIDs.GetSize();
		for (int i = 0; i < nCount; i++)
			if (pTBI->nItemIDs[i] == nID)
				return pTBI;
	}
	return NULL;
}

int CCommandMenu::CommandToIndex(CCommandMenu::ToolBarInfo* pTBI, UINT nID)
{
	const int nCount = pTBI->nItemIDs.GetSize();
	for (int i = 0; i < nCount; i++)
		if (pTBI->nItemIDs[i] == nID)
			return i;

	return -1;
}

void CCommandMenu::ReloadToolbarBitmaps()
{
	// Go to the top!
	CCommandMenu* pOwner = DYNAMIC_DOWNCAST(CCommandMenu, m_pOwner);
	if (pOwner)
	{
		pOwner->ReloadToolbarBitmaps();
		return;
	}

	POSITION pos = m_listToolBars.GetHeadPosition();
	while (pos)
	{
		ToolBarInfo* pTBI = m_listToolBars.GetNext(pos);
		pTBI->Reload();
	}
}

BOOL CCommandMenu::OnRefresh()
{
	// Recolour the toolbar bitmaps
	ReloadToolbarBitmaps();

	// Call base version
	return CImageMenu::OnRefresh();
}

CSize CCommandMenu::MeasureImage(LPMEASUREITEMSTRUCT lpMeasureItemStruct)
{
	CSize image_size = GetImageSize() + CSize(2 * g_cxIconSpace, 2 * g_cyIconSpace);
	CSize text_size = MeasureText(lpMeasureItemStruct);
	int size = max(text_size.cy, max(image_size.cx, image_size.cy));
	return CSize(size + 1, size);
}

void CCommandMenu::DrawImage(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	CDC& dc = *CDC::FromHandle(lpDrawItemStruct->hDC);
	CRect rcItem(lpDrawItemStruct->rcItem);
	UINT itemID = LOWORD(lpDrawItemStruct->itemID);
	BOOL fSelected = lpDrawItemStruct->itemState & ODS_SELECTED;
	BOOL fChecked = lpDrawItemStruct->itemState & ODS_CHECKED;
	BOOL fGrayed = lpDrawItemStruct->itemState & ODS_GRAYED;

	// Fill the background
	FillRect(dc, rcItem, GetSysColorBrush(COLOR_3DFACE));

	// Look for a corresponding toolbar button
	ToolBarInfo* pTBI = FindToolBar(itemID);
	if ((NULL == pTBI) && !fChecked)
	{
		// No icon or check
		int nColour = fSelected ? COLOR_HIGHLIGHT : COLOR_3DFACE;
		FillRect(dc, rcItem, GetSysColorBrush(nColour));
		return;
	}

	CRect rcOutside(rcItem);
	rcOutside.right -= 1; // Leave a pixel before the text
	CRect rcInside(rcOutside);
	rcInside.DeflateRect(1, 1);

	if (fChecked && (!fSelected || fGrayed))
	{
		// Chessboard pattern
		static WORD abPattern[8] = 
			{ 0x5555, 0xAAAA, 0x5555, 0xAAAA, 0x5555, 0xAAAA, 0x5555, 0xAAAA };
		CBitmap bmHatch; bmHatch.CreateBitmap(8, 8, 1, 1, &abPattern);
		CBrush brHatch; brHatch.CreatePatternBrush(&bmHatch);
		COLORREF crSave1 = dc.SetTextColor(GetSysColor(COLOR_3DHIGHLIGHT));
		COLORREF crSave2 = dc.SetBkColor(GetSysColor(COLOR_3DFACE));
		FillRect(dc, rcInside, brHatch);
		dc.SetTextColor(crSave1);
		dc.SetBkColor(crSave2);
	}

	if (pTBI)
	{
		int nIndex = CommandToIndex(pTBI, itemID);
		ASSERT(nIndex >= 0);

		// Centre icon
		CRect r(rcInside.CenterPoint(), pTBI->sizeImage);
		r.OffsetRect(-r.Width() / 2, -r.Height() / 2);

		if (fGrayed)
		{
			DimBlt(dc, r.left, r.top, r.Width(), r.Height(),
				pTBI->bmMono, nIndex * r.Width(), 0);
		}
		else
		{
			CDC dcMem; dcMem.CreateCompatibleDC(NULL);
			CBitmap* pbmOld = dcMem.SelectObject(&pTBI->bmNormal);
			TransBlt(dc, r.left, r.top, r.Width(), r.Height(),
				dcMem, nIndex * r.Width(), 0,
				pTBI->bmMask, nIndex * r.Width(), 0);
			dcMem.SelectObject(pbmOld);
		}
	}
	else if (fChecked)
	{
		// DrawFrameControl only likes drawing the check at its
		// 'natural size', so place it in the centre.
		CSize sizeCheck(
			GetSystemMetrics(SM_CXMENUCHECK), 
			GetSystemMetrics(SM_CYMENUCHECK));
		CRect r(rcInside.CenterPoint(), sizeCheck);
		r.OffsetRect(-r.Width() / 2, -r.Height() / 2);

		// Draw check into a mono bitmap (that's what DrawFrameControl does).
		CDC dcMem; dcMem.CreateCompatibleDC(NULL);
		CBitmap bm; bm.CreateBitmap(r.Width(), r.Height(), 1, 1, 0);
		CBitmap* pbmOld = dcMem.SelectObject(&bm);
		dcMem.DrawFrameControl(CRect(0, 0, r.Width(), r.Height()),
			DFC_MENU, DFCS_MENUCHECK);

		// Overlay check onto image.
		if (fGrayed)
		{
			dc.SetTextColor(GetSysColor(COLOR_3DHIGHLIGHT));
			TransBlt(dc, r.left + 1, r.top + 1, r.Width(), r.Height(), 
				dcMem, 0, 0, bm, 0, 0);
			dc.SetTextColor(GetSysColor(COLOR_3DSHADOW));
		}
		TransBlt(dc, r.left, r.top, r.Width(), r.Height(), 
			dcMem, 0, 0, bm, 0, 0);

		dcMem.SelectObject(pbmOld);
	}

	// Draw edge around image
	if (fChecked)
		dc.DrawEdge(rcOutside, BDR_SUNKENOUTER, BF_RECT);
	else if (fSelected && !fGrayed)
		dc.DrawEdge(rcOutside, BDR_RAISEDINNER, BF_RECT);
}

BOOL CCommandMenu::DestroyMenu()
{
	// Delete toolbar info
	while (!m_listToolBars.IsEmpty())
		delete m_listToolBars.RemoveHead();

	return CImageMenu::DestroyMenu();
}

CCommandMenu::~CCommandMenu()
{
	DestroyMenu();
}

//
// CIconMenu
//

IMPLEMENT_DYNCREATE(CIconMenu, CImageMenu)

CIconMenu::CIconMenu()
{
}

CIconMenu::~CIconMenu()
{
	HICON hicon;
	UINT nID;
	POSITION pos = m_mapIcons.GetStartPosition();
	while (pos)
	{
		m_mapIcons.GetNextAssoc(pos, nID, hicon);
		DestroyIcon(hicon);
	};
}

BOOL CIconMenu::SetIcon(int nIndex, HICON hicon)
{
	HICON hiconOld=0;
	if (m_mapIcons.Lookup(nIndex, hiconOld))
		DestroyIcon(hiconOld);
	m_mapIcons[nIndex] = hicon;

	// Make sure the item has been made owner-draw
	ProcessMenuItem(nIndex, TRUE);

	// Yawn
	MENUITEMINFO mii;
	ZeroMemory(&mii, sizeof(mii));
	mii.cbSize = sizeof(mii);
	mii.fMask = MIIM_DATA;
	VERIFY(GetMenuItemInfo(nIndex, TRUE, &mii));
	CItemData* pData = GetItemData(mii.dwItemData);
	if (NULL == pData)
	{
		// Item has not been made owner-draw yet
//		ASSERT(FALSE);
		return FALSE;
	}

	pData->m_hicon = hicon;
	return TRUE;
}

CImageMenu::CItemData* CIconMenu::CreateItemData(LPCTSTR lpItemText)
{
	CItemData* pData = new CItemData(this, lpItemText);
	m_listItemData.insert(pData);
	return pData;
}

CSize CIconMenu::MeasureImage(LPMEASUREITEMSTRUCT lpMeasureItemStruct)
{
//	TRACE("CIconMenu::MeasureImage: itemID = %x\n", lpMeasureItemStruct->itemID);
	return CSize(GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON))
		+ CSize(2 * g_cxIconSpace, 2 * g_cyIconSpace);
}

void CIconMenu::DrawImage(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	CDC& dc = *CDC::FromHandle(lpDrawItemStruct->hDC);
	CRect rcIcon(lpDrawItemStruct->rcItem);
	UINT itemID = LOWORD(lpDrawItemStruct->itemID);
	BOOL fSelected = lpDrawItemStruct->itemState & ODS_SELECTED;
	BOOL fChecked = lpDrawItemStruct->itemState & ODS_CHECKED;
	BOOL fGrayed = lpDrawItemStruct->itemState & ODS_GRAYED;

	// Fill the background
	FillRect(dc, rcIcon, GetSysColorBrush((fSelected)?COLOR_HIGHLIGHT:COLOR_3DFACE));

	HICON hicon;
	CItemData* pData = GetItemData(lpDrawItemStruct->itemData);
	if (pData && 0!=(hicon = pData->m_hicon))
	{
		// Adjust icon rectangle if necessary
		int cxSmIcon = GetSystemMetrics(SM_CXSMICON);
		int cySmIcon = GetSystemMetrics(SM_CYSMICON);
		if (rcIcon.Width() > cxSmIcon)
			rcIcon.DeflateRect((rcIcon.Width() - cxSmIcon) / 2, 0);
		if (rcIcon.Height() > cySmIcon)
			rcIcon.DeflateRect(0, (rcIcon.Height() - cySmIcon) / 2);

		DrawIconEx(dc, rcIcon.left, rcIcon.top, hicon, cxSmIcon, cySmIcon, 
			0, NULL, DI_NORMAL);
	}
}

