// ImageMenu.h
//
// Copyright (c) 1997, Jim Barry

#ifndef _IMAGEMENU_H_
#define _IMAGEMENU_H_

#include <afxtempl.h>
#include <shlobj.h>
#include <set>

// CImageMenu is a base class for self-draw menus. CImageMenu is
// fairly abstract, but contains code for drawing the text portion 
// of a menu and the submenu arrows, as well as support for keyboard
// mnemonics. It is left to derived classes to implement drawing the 
// menu item images.
// 
// To create an image menu from a menu resource, create a CImageMenu
// object and call LoadMenu as normal. To make all aspects of the menu
// work correctly you must call OnMenuChar, OnInitMenuPopup, OnMeasureItem, 
// and OnSysColorChange from the corresponding handlers in the menu's 
// owner window. This applies equally to CImageMenu-derived classes.
//
// Dynamic creation of menus is also possible using CreatePopupMenu.
//
class CImageMenu : public CMenu
{
	DECLARE_DYNCREATE(CImageMenu)

public:
// Construction
	CImageMenu();

	static void Register(CWnd* pWnd);
	static void Refresh();

	BOOL LoadMenu(LPCTSTR lpszResourceName);
	BOOL LoadMenu(UINT nIDResource);
	BOOL Attach(HMENU hMenu, BOOL fIsPopupMenu);
	virtual BOOL DestroyMenu();
	// Must supply owner for owner-draw popups to work properly
	BOOL CreatePopupMenu(CImageMenu* pOwner, BOOL fAutoDelete = FALSE); 
	CImageMenu* GetSubMenu(int nPos);
	void ShowImages(BOOL fShow);
	CImageMenu* GetRoot();

// Overrides
	virtual void DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct);
	virtual void MeasureItem(LPMEASUREITEMSTRUCT lpMeasureItemStruct);

// Helpers
	BOOL GetMenuItemInfo(UINT uItem, BOOL fByPosition, LPMENUITEMINFO lpmii)
		{ return ::GetMenuItemInfo(*this, uItem, fByPosition, lpmii); }
	BOOL SetMenuItemInfo(UINT uItem, BOOL fByPosition, LPMENUITEMINFO lpmii)
		{ return ::SetMenuItemInfo(*this, uItem, fByPosition, lpmii); }
	BOOL InsertMenuItem(UINT uItem, BOOL fByPosition, LPMENUITEMINFO lpmii)
		{ return ::InsertMenuItem(*this, uItem, fByPosition, lpmii); }
	UINT GetMenuDefaultItem(BOOL fByPosition, UINT gmdiFlags = NULL)
		{ return ::GetMenuDefaultItem(*this, fByPosition, gmdiFlags); }
	BOOL SetMenuDefaultItem(UINT uItem, BOOL fByPosition)
		{ return ::SetMenuDefaultItem(*this, uItem, fByPosition); }

	static BOOL FindMnemonic(CString str, UINT nChar);
	static void AdjustABCWidths(CDC& dc, const CString& str, CRect& rect, BOOL fShrink);
	static void DrawMenuText(CDC& dc, const CString& strItem, const CRect& rect);

	// Derive from this to add item data
	class CID
	{
	public:
		enum { magic_number = 0x12345678 };
		CID(CImageMenu* pMenu, LPCTSTR lpszText)
			: m_id(magic_number), m_pMenu(pMenu), m_strText(lpszText) {}
		virtual ~CID() {}

		UINT m_id;
		CImageMenu* m_pMenu; // Owner
		CString m_strText; // Menu item text
	};
	typedef CID CItemData;

// Overridables
protected:
	// Measurement handlers for the menu elements
	virtual CSize MeasureImage(LPMEASUREITEMSTRUCT lpMeasureItemStruct);
	virtual CSize MeasureText(LPMEASUREITEMSTRUCT lpMeasureItemStruct);
	virtual CSize MeasureArrow(LPMEASUREITEMSTRUCT lpMeasureItemStruct);
	virtual CSize MeasureSeparator(LPMEASUREITEMSTRUCT lpMeasureItemStruct);

	// Draw handlers for the menu elements
	virtual void DrawImage(LPDRAWITEMSTRUCT lpDrawItemStruct);
	virtual void DrawText(LPDRAWITEMSTRUCT lpDrawItemStruct);
	virtual void DrawArrow(LPDRAWITEMSTRUCT lpDrawItemStruct);
	virtual void DrawSeparator(LPDRAWITEMSTRUCT lpDrawItemStruct);

	// Item data management
	virtual CItemData* CreateItemData(LPCTSTR lpItemText);
	virtual void DestroyItemData(CItemData* pData);
	static CItemData* GetItemData(DWORD dwItemData);
	static CString GetTextFromItemData(DWORD dwItemData);

// Implementation
public:
	// Forwarders
	static LRESULT OnMenuChar(UINT nChar, UINT nFlags, CMenu* pMenu);
	static void OnInitMenuPopup(CMenu* pPopupMenu, UINT nIndex, BOOL bSysMenu);
	static BOOL OnMeasureItem(int nIDCtl, LPMEASUREITEMSTRUCT lpMeasureItemStruct);
	static BOOL OnDrawItem(int nIDCtl, LPDRAWITEMSTRUCT lpDrawItemStruct);

public:
	CImageMenu(CImageMenu* pOwner);
	~CImageMenu();
	virtual BOOL OnRefresh();
	virtual void OnInitMenuPopup(UINT nIndex, BOOL bSysMenu);
	virtual LRESULT OnMenuChar(UINT nChar, UINT nFlags);
	virtual void OnMenuSelect(UINT uItem, UINT fuFlags, HMENU hmenu);
	virtual void OnMenuClose();
	virtual BOOL ProcessMenuItem(UINT nItem, BOOL fShowImages);
	virtual BOOL ProcessMenuItems(BOOL fShowImages);
	static CImageMenu* FromHandlePermanent(HMENU hMenu);

protected:
	enum Flag {
		show_images = 1,
		popup_menu = 2,
		destroy_hmenu = 4
	};
	static CFont s_fontMenu;
	static CFont s_fontMenuBold;
	CImageMenu* m_pOwner;
	int m_nImageWidth;
	int m_nArrowWidth;
	UINT m_flags;
	typedef std::set<CItemData*> ItemDataPool;
	ItemDataPool m_listItemData;
	typedef std::set<CImageMenu*> SubMenuPool;
	SubMenuPool m_listSubMenus;

	static CFont* GetFont();
	static CFont* GetBoldFont();
	BOOL GetFlag(Flag flag);
	void SetFlag(Flag flag, BOOL fSet);
	BOOL IsPopupMenu();
};

// CCommandMenu is a class which implements Office-style menus.
// The menu item images are extracted from toolbar resources, which
// are fed to the menu with the AddToolBar function.
class CCommandMenu : public CImageMenu
{
	typedef CImageMenu Base;
	DECLARE_DYNCREATE(CCommandMenu)

public:
	CCommandMenu();
	BOOL AddToolbar(LPCTSTR lpszResourceName);
	BOOL AddToolbar(UINT nID);
	BOOL AddToolbars(const UINT* id_list, UINT count);
	static void ShowButtons(BOOL fShow);
	static BOOL ButtonsVisible();

// Overrides
public:
	~CCommandMenu();
	virtual BOOL OnRefresh();
	virtual BOOL DestroyMenu();
	virtual void DrawImage(LPDRAWITEMSTRUCT lpDrawItemStruct);
	virtual CSize MeasureImage(LPMEASUREITEMSTRUCT lpMeasureItemStruct);

// Implementation
protected:
	struct ToolBarInfo; // forward decl
	ToolBarInfo* FindToolBar(UINT nID);
	int CommandToIndex(ToolBarInfo* pTBI, UINT nID);
	void ReloadToolbarBitmaps();
	CSize GetImageSize();

protected:
	struct ToolBarInfo
	{
		void Reload();
		HINSTANCE hInstImageWell;
		HRSRC hRsrcImageWell;
		CBitmap bmNormal;
		CBitmap bmMask;
		CBitmap bmMono;
		CSize sizeImage;
		CArray<UINT, UINT> nItemIDs;
	};
	CList<ToolBarInfo*, ToolBarInfo*> m_listToolBars;
	CSize m_sizeImage;
	static BOOL s_fShowButtons;
};

// CIconMenu is a class which shows a small icon by each menu item.
// Create the menu in the usual way then associate icons with items
// using SetIcon. This menu class is primarily designed for popup
// menus.
class CIconMenu : public CImageMenu
{
	DECLARE_DYNCREATE(CIconMenu)

public:
	CIconMenu();
	BOOL SetIcon(int nIndex, HICON hicon);

// Overrides
public:
	~CIconMenu();
	virtual void DrawImage(LPDRAWITEMSTRUCT lpDrawItemStruct);
	virtual CSize MeasureImage(LPMEASUREITEMSTRUCT lpMeasureItemStruct);
	virtual CImageMenu::CItemData* CreateItemData(LPCTSTR lpItemText);

// Implementation
protected:
	class CItemData : public CImageMenu::CItemData
	{
	public:
		CItemData(CImageMenu* pMenu, LPCTSTR lpszText)
			: CImageMenu::CID(pMenu, lpszText), m_hicon(NULL) {}
		HICON m_hicon;
	};
	static CItemData* GetItemData(DWORD dwItemData)
		{ return dynamic_cast<CItemData*>(CImageMenu::GetItemData(dwItemData)); }
	CMap<UINT, UINT, HICON, HICON> m_mapIcons;
};

#endif // _IMAGEMENU_H_
