/////////////////////////////////////////////////////////////////////////////
// Persist.h
//
//@doc

#pragma once
#ifndef INC_PERSIST_H
#define INC_PERSIST_H

// Library dependencies
#ifdef PERSIST_USE_LIB
	#ifdef _AFXDLL
		#ifdef _DEBUG
			#ifdef _UNICODE
				#pragma comment(lib, "persist_ud.lib")
			#else
				#pragma comment(lib, "persist_d.lib")
			#endif
		#else
			#ifdef _UNICODE
				#pragma comment(lib, "persist_u.lib")
			#else
				#pragma comment(lib, "persist.lib")
			#endif
		#endif
	#else
		#ifdef _DEBUG
			#ifdef _UNICODE
				#pragma comment(lib, "persist_usd.lib")
			#else
				#pragma comment(lib, "persist_sd.lib")
			#endif
		#else
			#ifdef _UNICODE
				#pragma comment(lib, "persist_us.lib")
			#else
				#pragma comment(lib, "persist_s.lib")
			#endif
		#endif
	#endif
#endif

/*
@class 
	The CPersistence class provides a framework for creating persistent objects. 
	There is built in support for strings, integers, points, rectangles, 
	and window positions. CPersistence is also designed to be extended to 
	accommodate other data types.
	
	<l CPersistence Overview>.
*/
class CPersistence
{
// @access Construction
public:
	//@cmember Constructs the object.
	CPersistence(LPCTSTR lpProfileSection);
	virtual ~CPersistence();

// @access Attributes
	//@cmember Changes the section where the persistent data is stored.
	void SetProfileSection(LPCTSTR lpSection);
	//@cmember Returns the current direction of the data exchange.
	BOOL IsStoring() { return m_fStoring; }
	//@cmember Sets the data exchange direction to read.
	void SetReadMode() { m_fStoring = FALSE; }
	//@cmember Sets the data exchange direction to write.
	void SetWriteMode() { m_fStoring = TRUE; }

// @access Operations
	//@cmember Exchanges the data in the current direction.
	BOOL ExchangePersistentData(BOOL fSave);
	//@cmember Sets the direction to read, and exchanges the data.
	BOOL LoadPersistentData() { return ExchangePersistentData(FALSE); }
	//@cmember Sets the direction to write, and exchanges the data.
	BOOL SavePersistentData() { return ExchangePersistentData(TRUE); }

// @access Overridables
	//@cmember Override this function to implement the data exchange.
	virtual void OnExchangePersistentData();

// @access Data exchange helpers
	//@cmember Exchanges text string data.
	void PDX_String(LPCTSTR lpKey, CString& str, LPCTSTR lpDefault = NULL);
	//@cmember Exchanges signed integer data.
	void PDX_Int(LPCTSTR lpKey, int& i, int nDefault = 0);
	//@cmember Exchanges unsigned integer data.
	void PDX_Uint(LPCTSTR lpKey, UINT& i, UINT nDefault = 0);
	//@cmember Exchanges unsigned short integer data.
	void PDX_Word(LPCTSTR lpKey, WORD& w, WORD wDefault);
	//@cmember Exchanges unsigned long integer data.
	void PDX_Dword(LPCTSTR lpKey, DWORD& w, DWORD wDefault);
	//@cmember Exchanges boolean data.
	void PDX_Bool(LPCTSTR lpKey, BOOL& f, BOOL fDefault);
	//@cmember Exchanges rectangle data.
	void PDX_Rect(LPCTSTR lpKey, RECT& rect, LPCRECT pDefault = NULL);
	//@cmember Exchanges point data.
	void PDX_Point(LPCTSTR lpKey, POINT& pt, LPPOINT pDefault = NULL);
	//@cmember Exchanges size data.
	void PDX_Size(LPCTSTR lpKey, SIZE& size, LPSIZE pDefault = NULL);
	//@cmember Exchanges window position data.
	void PDX_WindowPos(LPCTSTR lpKey, CWnd* pWnd, LPPOINT lpDefault = NULL);
	//@cmember Exchanges window size data.
	void PDX_WindowSize(LPCTSTR lpKey, CWnd* pWnd, LPSIZE lpDefault = NULL);
	//@cmember Exchanges window rectangle data.
	void PDX_WindowRect(LPCTSTR lpKey, CWnd* pWnd, LPCRECT lpDefault = NULL);
	//@cmember Exchanges window rectangle data for frame windows.
	void PDX_MainFrameRect(LPCTSTR lpKey, CFrameWnd* pWnd,
		LPCRECT lprcDefault = NULL);
	//@cmember Exchanges window rectangle data for frame windows.
	void PDX_MainFrameRect(LPCTSTR lpKey, CFrameWnd* pWnd,
		const SIZE sizeDefault);
	//@cmember Exchanges toolbar state data.
	void PDX_BarState(LPCTSTR lpKey, CFrameWnd* pWnd);
	//@cmember Exchanges vector of ints.
	void PDX_IntVec(LPCTSTR lpKey, int* pVec, UINT nCount);
	//@cmember Exchanges binary data.
	void PDX_Data(LPCTSTR lpKey, void* pData, UINT cbData);
	
// Implementation
protected:
	CString m_strProfileSection;
	BOOL m_fStoring;
};


/////////////////////////////////////////////////////////////////////////////
// CFileDialogX -	File Dialog that remembers its initial folder and 
//					filter index
//
// Stores the initial folder and filter index in profile entries:
//   Profile section: controlled by CPersistence
//   Profile entry for initial folder: "InitialFolder".
//   Profile entry for filter index: "FilterIndex".
//
// The profile section can be set by calling 
// CPersistence::SetProfileSection after construction.
// In all other respects CFileDialogX behaves identically to 
// CFileDialog.

class CFileDialogX : public CFileDialog, public CPersistence
{
	DECLARE_DYNAMIC(CFileDialogX)

public:
	CFileDialogX::CFileDialogX(
		BOOL fOpen, LPCTSTR lpszDefExt = NULL, 
		LPCTSTR lpszFileName = NULL,
		DWORD dwFlags = OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
		LPCTSTR lpszFilter = NULL, CWnd* pParentWnd = NULL);

	void SetDefaultInitialFolder(LPCTSTR lpszFolder);

// Overrides
	int DoModal();
	void OnExchangePersistentData();

// Implementation
private:
	CString m_strInitialFolder;
	int m_nFilterIndex;
};


/////////////////////////////////////////////////////////////////////////////
// CPropertySheetX - PropertySheet that remembers its page index
//
// Stores the index of the last active page in a profile entry:
//   Profile section: same as PropertySheet caption
//   Profile entry: "Last Page".
//
// The profile section can be changed by calling 
// CPersistence::SetProfileSection after construction.
// In all other respects CPropertySheetX behaves identically to 
// CPropertySheet.
// 
// Tested only for modal use - should be OK for modeless though.

class CPropertySheetX : public CPropertySheet, public CPersistence
{
	DECLARE_DYNAMIC(CPropertySheetX)

// Construction
public:
	CPropertySheetX(UINT nIDCaption, CWnd* pParentWnd = NULL, UINT iSelectPage = 0);
	CPropertySheetX(LPCTSTR pszCaption, CWnd* pParentWnd = NULL, UINT iSelectPage = 0);

// Overrides
public:
	virtual void OnExchangePersistentData();
	virtual BOOL DestroyWindow();
	virtual BOOL OnInitDialog();
	virtual BOOL OnNotify(WPARAM wParam, LPARAM lParam, LRESULT* pResult);

// Implementation
private:
	int m_nPage; // Recent page index
};

/////////////////////////////////////////////////////////////////////////////

#endif // INC_PERSIST_H